import contextlib
import datetime
import json
import logging
import uuid

import flask
import flask_security
import requests

# from authlib.oauth2.rfc6749 import OAuth2Token
from authlib.integrations.flask_client import OAuth
from flask_login import current_user
from mxcubecore import HardwareRepository as HWR
from mxcubecore.model.lims_session import LimsSessionManager

from mxcubeweb.core.components.component_base import ComponentBase
from mxcubeweb.core.models.usermodels import User
from mxcubeweb.core.util.networkutils import is_local_host

HTTP_REQUESTS_TIMEOUT = 10  # Generic timeout for HTTP requests in seconds


class BaseUserManager(ComponentBase):
    """Base class for managing user-related operations.

    Operation it manages are: authentication, session management, and Single Sign-On
    (SSO) integration. It provides methods to handle user login, logout, session
    updates, and role assignments. The class also includes functionality to manage
    active users, set operators, and validate SSO tokens. It is designed to be extended
    by more specific user manager implementations.

    Attributes:
        app: The application instance.
            It is used to access various components and configurations such as
            the Flask server instance and the configuration settings.
        config: The configuration settings for the user manager.
            It contains information such as SSO client ID, client secret and
            metadata URI.
    """

    def __init__(self, app, config):
        super().__init__(app, config)
        HWR.beamline.lims.connect("sessionsChanged", self.handle_sessions_changed)
        self.oauth_client = OAuth(app=app.server.flask)

        self.oauth_client.register(
            name="keycloak",
            client_id=self.app.CONFIG.sso.CLIENT_ID,
            client_secret=self.app.CONFIG.sso.CLIENT_SECRET,
            server_metadata_url=self.app.CONFIG.sso.META_DATA_URI,
            client_kwargs={
                "scope": "openid email profile",
                "code_challenge_method": "S256",  # enable PKCE
            },
        )

    def handle_sessions_changed(self, sessions):
        self.app.server.emit("sessionsChanged", namespace="/hwr")

    def get_observers(self) -> list[User]:
        """List users that are in observer mode.

        Observer mode means user is logged in (authenticated and active) but not in
        control of the application.
        """
        return [
            user
            for user in User.query.all()
            if ((not user.in_control) and user.is_authenticated and user.is_active)
        ]

    def get_operator(self) -> User:
        """Return user object that is controlling the beamline (operator)."""
        user = None

        for _u in User.query.all():
            if _u.in_control:
                user = _u
                break

        return user

    def is_operator(self) -> bool:
        """Check if current user is an operator.

        Returns:
            ``True`` if the current_user is an operator.
        """
        return getattr(current_user, "in_control", False)

    def active_logged_in_users(self, exclude_inhouse: bool = False) -> list[User]:
        """List of active and logged in users.

        Args:
            exclude_inhouse (bool): exclude inhouse users from the list
        """
        self.update_active_users()

        if exclude_inhouse:
            users = [
                _u.username for _u in User.query.all() if _u.active and not _u.isstaff
            ]
        else:
            users = [_u.username for _u in User.query.all() if _u.active]

        return users

    def get_user(self, username: str) -> User | None:
        """Return user model instance based on username."""
        user = None

        for _u in User.query.all():
            if _u.username == username:
                user = _u

        return user

    def set_operator(self, username: str) -> User | None:
        """Set the user with the given username to be an operator."""
        user = None

        for _u in User.query.all():
            if _u.username == username:
                self.db_set_in_control(_u, True)
                user = _u
            else:
                self.db_set_in_control(_u, False)

        return user

    def update_active_users(self) -> None:
        """Check if any user have been inactive for longer than session lifetime.

        If so, deactivate the user in datastore and emit the relvant signals
        ``userChanged`` and ``observersChanged`` to the client.
        """
        for _u in User.query.all():
            if (
                _u.active
                and _u.last_request_timestamp
                and (datetime.datetime.now() - _u.last_request_timestamp)
                > flask.current_app.permanent_session_lifetime
            ):
                logging.getLogger("MX3.HWR").info(
                    "Logged out inactive user %s", _u.username
                )
                self.app.server.user_datastore.delete_user(_u)
                self.app.server.user_datastore.commit()

                self.app.server.emit(
                    "userChanged", room=_u.socketio_session_id, namespace="/hwr"
                )

        self.app.server.emit("observersChanged", namespace="/hwr")

    def update_operator(self, new_login: bool = False) -> None:
        """Sets the operator based on the logged in users.

        If no user is currently in control, the first logged in user is set.
        Additionally, proposal is set based on the operator selected_proposal field.

        Args:
            new_login: True if method was invoked with new user login.
        """
        if not current_user.is_anonymous:
            active_in_control = False

            for _u in User.query.all():
                if _u.is_authenticated and _u.in_control:
                    active_in_control = True
                else:
                    self.db_set_in_control(_u, False)

            # If new login and new observer login, clear nickname
            # so that the user get an opertunity to set one
            if new_login:
                current_user.nickname = ""

            # If no user is currently in control set this user to be
            # in control
            if not active_in_control:
                if not HWR.beamline.lims.is_user_login_type():
                    # current_user.nickname = self.app.lims.get_proposal(current_user)
                    current_user.fullname = HWR.beamline.lims.get_full_user_name()
                    current_user.nickname = HWR.beamline.lims.get_user_name()
                else:
                    current_user.nickname = current_user.username

                self.db_set_in_control(current_user, True)

            # Set active proposal to that of the active user
            for _u in User.query.all():
                if _u.is_authenticated and _u.in_control:
                    if not HWR.beamline.lims.is_user_login_type():
                        # In principle there is no need for doing so..
                        self.app.lims.select_session(
                            self.app.lims.get_session_manager().active_session.proposal_name
                        )  # The username is the proposal
                    elif _u.selected_proposal is not None:
                        self.app.lims.select_session(_u.selected_proposal)

    def is_inhouse_user(self, user_id: str) -> bool:
        """Check if the ``user_id`` is in the in-house user list.

        Args:
            user_id: user id composed from code and number.

        Returns:
            ``True`` if ``user_id`` is in the in-house user list, ``False`` otherwise.
        """
        user_id_list = [
            "%s%s" % (code, number)
            for (code, number) in HWR.beamline.session.in_house_users
        ]

        return user_id in user_id_list

    # Abstract method to be implemented by concrete implementation
    def _login(self, login_id: str, password: str) -> LimsSessionManager:
        pass

    def sso_validate(self) -> str:
        try:
            token_response = self.oauth_client.keycloak.authorize_access_token()
            username = token_response["userinfo"]["preferred_username"]
            token = token_response["access_token"]
        except Exception as e:
            raise e
        else:
            self.login(username, token, sso_data=token_response)

    def sso_token_expired(self) -> bool:
        res = json.loads(
            requests.post(
                self.app.CONFIG.sso.TOKEN_INFO_URI,
                headers={"Authorization": "Bearer %s" % current_user.token},
                data={
                    "grant_type": "refresh_token",
                    "refresh_token": current_user.refresh_token,
                },
                timeout=HTTP_REQUESTS_TIMEOUT,
            ).json()
        )

        return "access_token" not in res

    def handle_sso_logout(self):
        if current_user.is_anonymous and self.sso_token_expired():
            self.signout()

    def login(self, login_id: str, password: str, sso_data: dict = {}) -> None:
        """Login the user.

        Create new session for the user if it does not exist. Activate user in
        data store. If a sample is loaded in sample changer but not mounted,
        mount it and update the smaple list. Try update the operator.

        Args:
            login_id: The username.
            password: The password.
            sso_data: Dictionary containing information from the SSO service used.
        """
        try:
            self._login(login_id, password)
        except Exception as e:
            self._signout(sso_data=sso_data)
            logging.getLogger("MX3.HWR").error(str(e))
            raise e
        else:
            if "sid" not in flask.session:
                flask.session["sid"] = str(uuid.uuid4())

            # Making sure that the session of any in active users are invalidated
            # before calling login
            self.update_active_users()

            user = self.db_create_user(login_id, password, sso_data)
            self.app.server.user_datastore.activate_user(user)
            flask_security.login_user(user, remember=False)

            # Important to make flask_security user tracking work
            self.app.server.security.datastore.commit()
            self.update_operator(new_login=True)

            msg = "User %s signed in" % user.username
            logging.getLogger("MX3.HWR").info(msg)

    # Abstract method to be implemented by concrete implementation
    def _signout(self):
        pass

    def signout(self) -> None:
        """Sign out the current user.

        If the user was an operator, the queue and samples are restored to init values,
        the session is cleared, the user is not an operator anymore. Log out and
        deactivte the user, and emit 'observersChanged' signal.
        """
        self._signout()
        user = current_user

        # If operator logs out clear queue and sample list
        if self.is_operator():
            if hasattr(HWR.beamline.session, "clear_session"):
                HWR.beamline.session.clear_session()

            self.db_set_in_control(current_user, False)

            msg = "User %s signed out" % user.username
            logging.getLogger("MX3.HWR").info(msg)

        self.app.server.user_datastore.delete_user(user)
        self.app.server.user_datastore.commit()
        flask_security.logout_user()

        self.app.server.emit("observersChanged", namespace="/hwr")

    def is_authenticated(self) -> bool:
        """Check if the current user is authenticated.

        Returns:
            ``True`` if the current user is authenticated.
        """
        return current_user.is_authenticated()

    def force_signout_user(self, username: str) -> None:
        """Force signout of the annonymous or non operating user.

        Args:
            username: username of the user to be signed out.
        """
        user = self.get_user(username)

        if not user.in_control or current_user.is_anonymous:
            socketio_sid = user.socketio_session_id
            HWR.beamline.lims.remove_user(username)
            self.app.server.user_datastore.delete_user(user)
            self.app.server.user_datastore.commit()
            self.app.server.emit("forceSignout", room=socketio_sid, namespace="/hwr")

    def login_info(self) -> dict:
        """Get the login information to be displayed in the application.

        Login information to be displayed in the application such as:
        * synchrotron and beamline names
        * user infromation
        * proposals list
        * selected proposal
        * and so on

        Returns:
            Dictionary with login information.
        """
        # update_operator will update the login status of current_user, and make
        # sure that the is_anonymous has the correct value.
        # Update operator calls lims.select_session that raises an exception if
        # there are no valid LIMS sessions.
        with contextlib.suppress(Exception):
            self.update_operator()

        login_type = "User" if HWR.beamline.lims.is_user_login_type() else "Proposal"

        if current_user.is_anonymous:
            self._signout()
            logging.getLogger("MX3.HWR").info("Logged out")
            return {
                "loggedIn": False,
                "useSSO": self.app.CONFIG.sso.USE_SSO,
                "loginType": login_type,
            }

        session_manager: LimsSessionManager = self.app.lims.get_session_manager()

        # If no previous session selected and a single session available
        # then it selects automatically the session
        if (
            current_user.selected_proposal is None
            and session_manager.active_session is not None
        ):
            self.app.lims.select_session(session_manager.active_session.session_id)

        res = {
            "synchrotronName": HWR.beamline.session.synchrotron_name,
            "beamlineName": HWR.beamline.session.beamline_name,
            "loggedIn": True,
            "loginType": login_type,
            "limsName": [item.dict() for item in HWR.beamline.lims.get_lims_name()],
            "proposalList": [session.__dict__ for session in session_manager.sessions],
            "rootPath": HWR.beamline.session.get_base_image_directory(),
            "user": current_user.todict(),
            "useSSO": self.app.CONFIG.sso.USE_SSO,
            "sessionRefreshInterval": self.app.CONFIG.app.SESSION_REFRESH_INTERVAL,
        }

        res["selectedProposal"] = "%s%s" % (
            HWR.beamline.session.proposal_code,
            HWR.beamline.session.proposal_number,
        )
        res["selectedProposalID"] = HWR.beamline.session.proposal_id

        return res

    def update_user(self, user: User) -> None:
        """Update user information in datastore.

        Args:
            user: User model instance.
        """
        self.app.server.user_datastore.put(user)
        self.app.server.user_datastore.commit()

    def _get_configured_roles(self, user: str) -> list[str]:
        """Get the roles configured for the user.

        Inhouse user has always assigned staff role additionaly.

        Args:
            user: username.
        """
        roles = set()

        _ihs = ["%s%s" % prop for prop in HWR.beamline.session.in_house_users]

        if self.config.inhouse_is_staff and user in _ihs:
            roles.add("staff")

        for _u in self.config.users:
            if _u.username == user:
                roles.add(_u.role)
                break

        return list(roles)

    def db_create_user(self, user: str, password: str, sso_data: dict) -> User:
        """Create or update user in datastore.

        If the user already exists, update the user information. If not create new one.
        Assign roles to the user, prevoiusly making sure the roles of 'staff' and
        'incontrol' existis in data store. If not create them also.

        Args:
            user: representation of username (eventually part of it).
                Also a nickname for new users.
            password: password (unused).
            sso_data: dictionary containing information from the SSO service used.

        Returns:
            User model instance existing in or added to datastore.
        """
        sid = flask.session["sid"]
        user_datastore = self.app.server.user_datastore

        username = HWR.beamline.lims.get_user_name()
        fullname = HWR.beamline.lims.get_full_user_name()

        # Make sure that the roles staff and incontrol always exists
        if not user_datastore.find_role("staff"):
            user_datastore.create_role(name="staff")
            user_datastore.create_role(name="incontrol")
            self.app.server.user_datastore.commit()

        _u = user_datastore.find_user(username=username)

        if not _u:
            if not HWR.beamline.lims.is_user_login_type():
                selected_proposal = user
            else:
                selected_proposal = None

            user_datastore.create_user(
                username=username,
                fullname=fullname,
                password="",
                nickname=user,
                session_id=sid,
                selected_proposal=selected_proposal,
                refresh_token=sso_data.get("refresh_token", str(uuid.uuid4())),
                token=sso_data.get("token", str(uuid.uuid4())),
                roles=self._get_configured_roles(user),
            )
        else:
            _u.refresh_token = sso_data.get("refresh_token", str(uuid.uuid4()))
            _u.token = sso_data.get("token", str(uuid.uuid4()))
            user_datastore.append_roles(_u, self._get_configured_roles(user))

        self.app.server.user_datastore.commit()

        return user_datastore.find_user(username=username)

    def db_set_in_control(self, user: User, control: bool) -> None:
        """Update users (their in_control field) in the datastore.

        If the passed user becomes an operator (``control=True``), the remaining users'
        in_control fields are set to ``False``. If passed user stops being an operator,
        only its in_control field is set to ``False``.

        Args:
            user: User model instance.
            control: the user becomes an operator (``True``) or not (``False``).
        """
        user_datastore = self.app.server.user_datastore

        if control:
            for _u in User.query.all():
                if _u.username == user.username:
                    _u.in_control = True
                else:
                    _u.in_control = False

                user_datastore.put(_u)
        else:
            _u = user_datastore.find_user(username=user.username)
            _u.in_control = control
            user_datastore.put(_u)

        self.app.server.user_datastore.commit()


class UserManager(BaseUserManager):
    """Class to provide specific implementations for user login and signout operations.

    It includes methods to handle login conditions such as checking if the user is
    active, anonymous, in-house, or accessing locally/remotely. The class also ensures
    that only one user can be logged in at a time and restricts in-house logins
    to local hosts. Additionally, it handles Single Sign-On (SSO) logout by making
    a request to the configured SSO logout URI.
    """

    def __init__(self, app, config):
        super().__init__(app, config)

    def _debug(self, msg: str):
        logging.getLogger("HWR").debug(msg)

    def _login(self, login_id: str, password: str) -> LimsSessionManager:
        """Check loging conditions such as: active, anonymous, inhouse, local/remote.

        Args:
            login_id: username
            password: password

        Returns:
            LimsSessionManager object with login information.
        """
        self._debug("_login. login_id=%s" % login_id)
        local_domains = self.app.CONFIG.app.LOCAL_DOMAINS
        try:
            session_manager: LimsSessionManager = HWR.beamline.lims.login(
                login_id, password, is_local_host(local_domains)
            )
        except Exception as e:
            logging.getLogger("MX3.HWR").error(e)
            raise e

        self._debug(
            "_login. proposal_tuple retrieved. Sessions=%s "
            % str(len(session_manager.sessions))
        )

        if login_id in self.active_logged_in_users():
            if current_user.is_anonymous:
                self.force_signout_user(login_id)
            else:
                if current_user.username == login_id:
                    msg = "You are already logged in"
                    raise RuntimeError(msg)
                msg = (
                    "Login rejected, you are already logged in"
                    " somewhere else\nand Another user is already"
                    " logged in"
                )
                raise RuntimeError(msg)

        # Only allow in-house log-in from local host
        if self.is_inhouse_user(login_id) and not is_local_host(local_domains):
            msg = "In-house only allowed from localhost"
            raise PermissionError(msg)

        # Only allow local login when remote is disabled
        if not self.app.ALLOW_REMOTE and not is_local_host(local_domains):
            msg = "Remote access disabled"
            raise ConnectionRefusedError(msg)

        return session_manager

    def _signout(self, sso_data=None):
        sso_data = sso_data or {}

        if self.app.CONFIG.sso.LOGOUT_URI:
            if not current_user.is_anonymous:
                HWR.beamline.lims.remove_user(current_user.username)
                refresh_token = current_user.refresh_token
            else:
                refresh_token = sso_data.get("refresh_token", None)

            requests.post(
                self.app.CONFIG.sso.LOGOUT_URI,
                data={
                    "client_id": self.app.CONFIG.sso.CLIENT_ID,
                    "client_secret": self.app.CONFIG.sso.CLIENT_SECRET,
                    "refresh_token": refresh_token,
                },
                timeout=HTTP_REQUESTS_TIMEOUT,
            )


class SSOUserManager(BaseUserManager):
    def __init__(self, app, config):
        super().__init__(app, config)

    def _login(self, login_id: str, password: str, sso: bool):
        return {"status": {"code": "ok", "msg": ""}}

    def _signout(self):
        pass



import contextlib
import logging
import re
from collections.abc import Callable
from functools import reduce
from typing import ClassVar

from flask import Blueprint, Response, jsonify, request
from pydantic import (
    BaseModel,
    ValidationError,
)

from mxcubeweb.core.server.openapidoc import OpenAPISpec


def valid_object_id(object_id: str) -> bool:
    """Validate that ``object_id`` contains only ``A-Z``, ``a-z``, and ``.`` , ``_``.

    Args:
        object_id (str): The string to validate.

    Returns:
        bool: ``True`` if the string is valid, ``False`` otherwise.
    """
    return bool(re.fullmatch(r"[A-Za-z._]+", object_id))


def validate_input_str(input_string: str) -> bool:
    """Validate that input string contains only alphanumeric characters and/or dot (.).

    Args:
        input_string (str): The string to validate.

    Returns:
        bool: True if the string is valid, False otherwise.
    """
    pattern = r"^[a-zA-Z0-9._]*$"
    return bool(re.match(pattern, input_string))


def assert_valid_type_arguments(func):
    """Make sure that all the arguments of the function are typehinted correctly.

    Make sure that all the arguments of the function are typehinted as
    pydantic model, float, int, str or bool
    """
    annotations = func.__annotations__

    # Loop through annotations and validate parameters from the request
    for param_name, param_type in annotations.items():
        if param_name == "return":
            continue

        # Raise RuntimerError If it's not a Pydantic model, float, int, str or bool
        if not issubclass(param_type, BaseModel | float | int | str | bool):
            msg = f"Argument {param_name} of {func} are not a pydantic model, float, \
                int or str"
            raise TypeError(msg)


log = logging.getLogger("MX3.HWR")


class ResourceHandlerFactory:
    _handlers: ClassVar[dict] = {}

    @classmethod
    def create_or_get(  # noqa: PLR0913
        cls,
        name: str,
        url_prefix: str,
        handler_dict: dict[str, object],
        app: object,
        exports: list[dict[str, str]],
        commands: list[str],
        attributes: list[str],
        handler_type="adapter",
    ) -> object:
        """Return or create handler.

        Returns:
            Existing handler if it exists, otherwise create and register a new one.
        """
        if name in cls._handlers:
            return cls._handlers[name]

        if handler_type == "component":
            resource_handler_cls = ComponentResourceHandler
        else:
            resource_handler_cls = AdapterResourceHandler

        handler = resource_handler_cls(
            name, url_prefix, handler_dict, app, exports, commands, attributes
        )
        cls._handlers[name] = handler
        return handler

    @classmethod
    def get_handler(cls, name: str) -> object | None:
        return cls._handlers.get(name)

    @classmethod
    def unregister(cls, name: str) -> bool:
        if name in cls._handlers:
            del cls._handlers[name]
            return True
        return False

    @classmethod
    def unregister_all(cls) -> bool:
        for key in cls.all_handlers():
            cls.unregister(key)

    @classmethod
    def all_handlers(cls) -> dict[str, object]:
        return cls._handlers.copy()

    @classmethod
    def register_with_server(cls, flask_server):
        for rh in cls.all_handlers().values():
            rh.register_blueprint(flask_server)


class ResourceHandler:
    openapi_spec = OpenAPISpec("docs", "/apidocs", "1.0.0", "MXCuBE Adapter API")

    def __init__(  # noqa: PLR0913
        self,
        name: str,
        url_prefix: str,
        handler_dict: dict[str, object],
        app: object,
        exports: list[dict[str, str]],
        commands: list[str],
        attributes: list[str],
    ) -> None:
        """Initialize the AdapterResourceHandler.

        Args:
            name: Name of the blueprint.
            url_prefix: URL prefix for the blueprint.
            handler_dict: Dictionary mapping object IDs to adapter objects.
            app: mxcube app object, providing acccess to mxcubecore and server
            exports: Predefined list of exported commands/attributes.
            commands: List of command names to export.
            attributes: List of attribute names to export.
        """
        self._bp = Blueprint(name, name, url_prefix=url_prefix)
        self._handler_dict = handler_dict
        self._server = app.server  # Store the server object to access its decorators
        self._app = app
        self._url_prefix = url_prefix
        self._exports = exports.copy()

        # Add export definitions for attributes and commands
        self._add_attribute_exports(attributes)
        self._add_command_exports(commands)

    @property
    def exports(self):
        return self._exports

    @property
    def commands(self):
        return [export for export in self._exports if export["method"] == "PUT"]

    @property
    def attributes(self):
        return [export for export in self._exports if export["method"] == "GET"]

    def _create_routes_for_exports(self) -> None:
        """Create Flask routes dynamically for each exported command or attribute."""
        for export in self._exports:
            route = self._get_handler_object_route(export)
            # For the time being we enforce the usage of pydantic models, int, float or
            # str for arguments to ensure safe validation of input. We rely on that
            # the pydantic models used are well specified. We validate the actual data
            # later.
            self._assert_valid_type_arguments(export)

            # Create the api doc before the view functions are created
            self._create_openapi_doc_for_view(route, export)

            # Create the view function dynamically
            view_func = self._create_view_func(export)

            # Apply decorators to the view function
            view_func = self._apply_decorators(view_func, export["decorators"])

            # Register route
            self._bp.add_url_rule(
                route,
                view_func=view_func,
                methods=[export["method"]],
                endpoint=export["attr"],
            )
            msg = (
                f"Registerd {route} to blueprint '{self._bp.name}' ({self._url_prefix})"
            )
            log.debug(msg)

    def _apply_decorators(
        self, view_func: Callable, decorators: list[Callable]
    ) -> Callable:
        """Applies a list of decorators to a view function.

        Args:
            view_func (Callable): The view function.
            decorators (List[Callable]): List of decorators.

        Returns:
            Callable: Decorated function.
        """
        return reduce(lambda f, decorator: decorator(f), decorators, view_func)

    def _create_view_func(self, export: dict[str, str]) -> Callable:
        """Creates a Flask view function for handling requests dynamically.

        Args:
            route (str): URL route.
            export (dict): Export definition with method, attr, and decorators.

        Returns:
            Callable: The view function.
        """

    def _validate_params_and_call_handler_func(
        self,
        export: dict[str, str],
        handler_obj: object,
    ) -> dict | Response:
        """Validates parameters from the request and calls the handler function.

        Args:
            export: Export definition with method, attr, and decorators.
            handler_obj: The handler object to call.

        Returns:
            The result of the handler function call or an error response.
        """
        # Get the method and its annotations
        handler_func = getattr(handler_obj, export["attr"])
        annotations = handler_func.__annotations__
        http_method = export["method"]

        # Prepare data for all required arguments
        validated_data = {}
        param_data = self._extract_param_data(http_method)

        # Validate parameters from the request
        for param_name, param_type in annotations.items():
            if param_name == "return":  # Skip the return annotation
                continue

            try:
                if param_data is not None:
                    validated_data[param_name] = self._validate_param_data(
                        param_name, param_type, param_data
                    )
            except (ValueError, TypeError) as ex:
                log.error(str(ex))  # noqa: TRY400
                msg = f"Invalid input for '{param_name}'"
                return (
                    jsonify({"error": msg}),
                    400,
                )

        # Call the view function with validated data
        try:
            result = handler_func(**validated_data)
        except Exception:
            msg = "Exception raised when calling view function"
            log.exception(msg)
            return jsonify({"error": msg}), 500
        else:
            # Handle and serialize the result
            return self._handle_view_result(result)

    def _assert_valid_type_arguments(self, export):
        """Ensure the method referenced in the export uses Pydantic arguments."""
        obj = next(iter(self._handler_dict.values()))
        assert_valid_type_arguments(getattr(obj, export["attr"]))

    def _create_openapi_doc_for_view(self, route, export):
        """Add OpenAPI documentation for a route."""
        # Get the first adapter object, the signature are all the same (same class) so
        # any will do for documentation purpose
        http_method = export["method"]
        obj = next(iter(self._handler_dict.values()))
        view_func = getattr(obj, export["attr"])
        annotations = view_func.__annotations__

        # Add OpenAPI documentation root for route
        self.openapi_spec.add_openapi_path(
            self._url_prefix, route, export, http_method, view_func
        )

        # Loop through annotations add response and arguments to OpenAPI spec for route
        for param_name, param_type in annotations.items():
            if param_name == "return":
                self.openapi_spec.add_openapi_response(
                    self._url_prefix, route, http_method, param_type
                )
                continue

            # If it's a Pydantic model, document
            if issubclass(param_type, BaseModel):
                self.openapi_spec.add_openapi_schema(
                    self._url_prefix, route, http_method, param_type
                )

    def _extract_param_data(self, http_method) -> dict:
        """Extract parameter data from request (JSON, query params, or form).

        Returns:
            Extracted data
        """
        # Prioritize JSON body, then query params, then form data
        # We are not really using query or form data, but they are added for
        # completness
        if http_method == "GET":
            return request.args
        return request.json or request.args.to_dict() or request.form.to_dict()

    def _validate_param_data(
        self, param_name: str, param_type: type, param_data: any
    ) -> dict | int | float | bool | str:
        """Validate a single parameter based on its expected type.

        Raises:
            ValueError: If validation fails or type is unsupported.
        """
        # If it's a Pydantic model, validate it
        if issubclass(param_type, BaseModel):
            try:
                # We only handle a single argument when using Pydantic models
                # We handle complex structures by checking if param_data
                # contains a value or a dictionary.
                if isinstance(param_data.get(param_name), dict):
                    return param_type.parse_obj(param_data[param_name])

                # If its a single value and Pydantic model, pass the entire
                # Pydantic model
                return param_type.parse_obj(param_data)
            except ValidationError as e:
                msg = f"Invalid input for '{param_name}' '{e.errors()}' with data"
                msg += f" '{param_data}' on: {request.url}"
                raise ValueError(msg) from e

        # If its not a Pydantic model, we expect the parameter to be present
        # and match the type hint (method signature)
        elif param_data is None or param_name not in param_data:
            msg = f"Missing required parameter '{param_name}' on: {request.url}"
            raise ValueError(msg)

        if isinstance(param_data[param_name], int | float | bool):
            # We consider int, float and bool safe and limits handled
            # by adapter or HardwareObject
            return param_data[param_name]
        if isinstance(param_data[param_name], str):
            # We consider str safe if it contains, alpha numerical
            # characters and dot "." and underscore "_"
            if validate_input_str(param_data[param_name]):
                return param_data[param_name]

            msg = f"Invalid string input for '{param_name}' '{param_data[param_name]}'"
            msg += f" on: {request.url}"
            raise ValueError(msg)

        msg = f"No model defined for '{param_name}'"
        msg += f" on: {request.url}"
        raise TypeError(msg)

    def _handle_view_result(self, result: object) -> dict | Response:
        """Handle the result of a view function.

        Handle the result of a view function,
        ensuring that it is serializable and properly formatted.

        Returns:
            Flask Response: JSON response.
        """
        try:
            if isinstance(result, Response):
                # If it's already a Flask Response, return it directly
                return result
            # Check if the result is a Pydantic model or any other serializable object
            if isinstance(result, BaseModel):
                # Convert Pydantic model to a dict
                result = result.dict()
            elif isinstance(result, dict):
                # If it's already a dictionary, it's ready for JSON serialization
                pass
            elif hasattr(result, "__dict__"):
                # If the result has __dict__ attribute (e.g., an object), convert to
                # dict
                result = result.__dict__
            elif isinstance(result, str | int | float | bool | list | tuple):
                # If it's already a simple type, no conversion needed
                result = {"return": result}
            else:
                return (
                    jsonify(
                        {
                            "error": (
                                f"Return value of type '{type(result)}' is not"
                                " serializable"
                            )
                        }
                    ),
                    500,
                )

            # Return the result as JSON (mime-type: application/json, code: 200)
            return jsonify(result)
        except Exception:
            msg = "An error occurred while processing the response."
            log.exception(msg)
            return (
                jsonify(
                    {
                        "error": msg,
                    }
                ),
                500,
            )

    def _add_exports(self, items: list[str], http_method: str) -> None:
        """Add export definitions to the EXPORTS list.

        Args:
            items: The list of commands or properties to add
            http_method: The HTTP method to use, GET, POST, PUT, DELETE
        """
        if http_method == "GET":
            decorators = [self._server.restrict]
        else:
            decorators = [self._server.require_control, self._server.restrict]

        for item in items:
            export = {
                "attr": item,
                "method": http_method,
                "decorators": decorators,
            }

            if not self.is_unique_export(export):
                msg = f"Export {export} already exists for {self._url_prefix}"
                raise ValueError(msg)

            self._exports.append(export)

    def is_unique_export(self, new_export):
        """Check if an export is unique.

        Check if an export with the same 'attr' and 'method' already exists in the
        EXPORTS list.

        Args:
            new_export (dict): The new export to check.

        Returns:
            bool: True if unique (no duplicates), False if a duplicate exists.
        """
        for export in self._exports:
            if (
                export["attr"] == new_export["attr"]
                and export["method"] == new_export["method"]
            ):
                return False
        return True

    def _add_command_exports(self, command_list) -> None:
        """Add PUT exports for commands."""
        self._add_exports(command_list, "PUT")

    def _add_attribute_exports(self, attribute_list) -> None:
        """Add GET exports for attributes."""
        self._add_exports(attribute_list, "GET")

    def register_blueprint(self, parent_bp) -> None:
        """Register the blueprint on the Flask server (server.flask).

        This allows the routes defined in the blueprint to be accessible on the server.
        """
        self._create_routes_for_exports()
        parent_bp.register_blueprint(self._bp)

        # Using try-except ot only register the documentation endpoint once
        with contextlib.suppress(ValueError):
            parent_bp.register_blueprint(self.openapi_spec.bp)

        msg = (
            f"Blueprint '{self._bp.name}' ({self._url_prefix}) registered with server."
        )
        log.debug(msg)


class ComponentResourceHandler(ResourceHandler):
    """Flask resource handler that dynamically creates routes for hardware objects.

    AdapterResourceHandler is a Flask resource handler that dynamically creates routes
    for hardware objects based on their attributes and methods.
    It supports GET and PUT requests for attributes and commands respectively.
    """

    def __init__(  # noqa: PLR0913
        self,
        name: str,
        url_prefix: str,
        handler_dict: dict[str, object],
        app: object,
        exports: list[dict[str, str]],
        commands: list[str],
        attributes: list[str],
    ) -> None:
        super().__init__(
            name, url_prefix, handler_dict, app, exports, commands, attributes
        )

    def _get_handler_object_route(self, export) -> str:
        """Return the base route for the resource handler."""
        return f"{export['url']}"

    def _create_view_func(self, export: dict[str, str]) -> Callable:
        """Create a Flask view function for handling requests dynamically.

        Args:
            route (str): URL route.
            export (dict): Export definition with method, attr, and decorators.

        Returns:
            Callable: The view function.
        """

        def _view_func(*args, **kwargs) -> any:  # noqa: ARG001
            # Get component from handler_dict, there is only one handler object for
            # components
            component_obj = next(iter(self._handler_dict.values()))
            component_name = next(iter(self._handler_dict.keys()))

            # We ensure that the component has the desired method
            if not hasattr(component_obj, export["attr"]):
                return (
                    jsonify(
                        {
                            "error": (
                                f"Method '{export['attr']}' not found on object"
                                f" '{component_name}'"
                            )
                        }
                    ),
                    404,
                )

            return self._validate_params_and_call_handler_func(export, component_obj)

        return _view_func


class AdapterResourceHandler(ResourceHandler):
    """Flask resource handler that dynamically creates routes for hardware objects.

    AdapterResourceHandler is a Flask resource handler that dynamically creates routes
    for hardware objects based on their attributes and methods.
    It supports GET and PUT requests for attributes and commands respectively.
    """

    def __init__(  # noqa: PLR0913
        self,
        name: str,
        url_prefix: str,
        handler_dict: dict[str, object],
        app: object,
        exports: list[dict[str, str]],
        commands: list[str],
        attributes: list[str],
    ) -> None:
        super().__init__(
            name, url_prefix, handler_dict, app, exports, commands, attributes
        )

    def _get_handler_object_route(self, export) -> str:
        """Return the base route for the resource handler."""
        # Dynamic route for handler object with id object_id, i.e:
        # /<object_id>/set_value
        return f"/<string:object_id>/{export['attr']}"

    def _create_view_func(self, export: dict[str, str]) -> Callable:
        """Create a Flask view function for handling requests dynamically.

        Args:
            route (str): URL route.
            export (dict): Export definition with method, attr, and decorators.

        Returns:
            Callable: The view function.
        """

        def _view_func(object_id: str, *args, **kwargs) -> any:  # noqa: ARG001
            # Validate object id
            if not valid_object_id(object_id):
                msg = f"Invalid object id '{object_id}'"
                log.error(msg)
                return jsonify({"error": msg}), 400

            # Check if the object_id exists in the handler_dict and corresponds to a
            # HardwareObject
            obj = self._app.mxcubecore.get_adapter(object_id)

            if not obj:
                msg = f"Object '{object_id}' not found"
                log.error(msg)
                return jsonify({"error": msg}), 404

            # We ensure that the object has the desired method
            if not hasattr(obj, export["attr"]):
                return (
                    jsonify(
                        {
                            "error": (
                                f"Method '{export['attr']}' not found on object"
                                f" '{object_id}'"
                            )
                        }
                    ),
                    404,
                )

            return self._validate_params_and_call_handler_func(export, obj)

        return _view_func



from typing import ClassVar

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract import AbstractEnergy

from mxcubeweb.core.adapter.actuator_adapter import ActuatorAdapter
from mxcubeweb.core.adapter.wavelength_adapter import WavelengthAdapter
from mxcubeweb.core.models.configmodels import ResourceHandlerConfigModel

resource_handler_config = ResourceHandlerConfigModel(
    commands=["set_value", "stop", "get_resolution_limits_for_energy"],
    attributes=["data", "get_value"],
)


class EnergyAdapter(ActuatorAdapter):
    """Adapter for Energy Hardware Object.

    A web socket is used to communicate information on longer running processes.
    """

    SUPPORTED_TYPES: ClassVar[list[object]] = [AbstractEnergy.AbstractEnergy]

    def __init__(  # noqa: D417
        self, ho, role, app
    ):
        """Initialize.

        Args:
            ho (object): Hardware object.
        """
        super().__init__(ho, role, app, resource_handler_config)
        self._add_adapter("wavelength", self._ho, WavelengthAdapter)
        self._type = "ENERGY"

    def get_resolution_limits_for_energy(self, energy: float) -> tuple:
        return HWR.beamline.resolution.get_limits_for_wavelength(
            self._ho.calculate_wavelength(energy)
        )



import json
import logging

from mxcubeweb import logging_handler
from mxcubeweb.core.components.component_base import ComponentBase
from mxcubeweb.core.models.adaptermodels import FrontEndStackTraceModel
from mxcubeweb.core.models.configmodels import (
    ResourceHandlerConfigModel,
)

hwr_logger = logging.getLogger("MX3.HWR")


class Log(ComponentBase):
    def __init__(self, app, config):
        super().__init__(
            app,
            config,
            resource_handler_config=ResourceHandlerConfigModel(
                exports=[
                    {
                        "attr": "log",
                        "method": "GET",
                        "url": "/",
                        "decorators": [app.server.restrict],
                    },
                    {
                        "attr": "log_frontend_traceback",
                        "method": "POST",
                        "url": "/log_frontend_traceback",
                        "decorators": [app.server.restrict],
                    },
                ]
            ),
        )

    def log(self):
        """Retrieve log messages."""
        messages = []

        for handler in logging.getLogger("MX3.HWR").handlers:
            if isinstance(handler, logging_handler.MX3LoggingHandler):
                messages = handler.buffer

        return messages

    def log_frontend_traceback(self, trace: FrontEndStackTraceModel) -> dict:
        """Log a UI traceback to the UI logger."""
        logging.getLogger("MX3.UI").error("------ Start of UI trace back ------")
        logging.getLogger("MX3.UI").error("Traceback: %s", trace.stack)
        logging.getLogger("MX3.UI").error(
            "State: %s", json.dumps(trace.state, indent=4)
        )
        logging.getLogger("MX3.UI").error("------ End of UI trace back ------")
        return {}



import logging
from typing import ClassVar

from mxcubecore.HardwareObjects.abstract import AbstractActuator, AbstractFlux

from mxcubeweb.core.adapter.adapter_base import ActuatorAdapterBase
from mxcubeweb.core.models.adaptermodels import (
    FloatValueModel,
    HOActuatorValueChangeModel,
)
from mxcubeweb.core.models.configmodels import ResourceHandlerConfigModel
from mxcubeweb.core.util.networkutils import RateLimited

resource_handler_config = ResourceHandlerConfigModel(
    commands=["set_value", "stop"], attributes=["data", "get_value"]
)

hwr_logger = logging.getLogger("MX3.HWR")


class ActuatorAdapter(ActuatorAdapterBase):
    """Adapter for Energy Hardware Object.

    A web socket is used to communicate information on longer running processes.
    """

    SUPPORTED_TYPES: ClassVar[list[object]] = [
        AbstractActuator.AbstractActuator,
        AbstractFlux.AbstractFlux,
    ]

    def __init__(  # noqa: D417
        self,
        ho,
        role,
        app,
        resource_handler_config=resource_handler_config,
    ):
        """Initialize.

        Args:
            ho (object): Hardware object.
        """
        super().__init__(ho, role, app, resource_handler_config)
        self._event_rate = 4

        @RateLimited(self._event_rate)
        def _vc(value, **kwargs):
            self.value_change(value, **kwargs)

        self._vc = _vc

        try:
            ho.connect("valueChanged", self._value_change)
            ho.connect("stateChanged", self.state_change)
        except Exception:
            msg = f"Could not connect to valueChanged or stateChanged: {ho.name}"
            hwr_logger.exception(msg)

    def _value_change(self, *args, **kwargs):
        self._vc(*args, **kwargs)

    def set_value(self, value: HOActuatorValueChangeModel) -> str:
        """Execute the sequence to set the value.

        Args:
            value (float): Target energy [keV].

        Returns:
            (float as str): The actual value set.

        Raises:
            ValueError: Value not valid or attemp to set a non-tunable energy.
            RuntimeError: Timeout while setting the value.
            StopItteration: When a value change was interrupted (abort/cancel).
        """
        self._ho.set_value(float(value.value))
        return self.get_value()

    def get_value(self) -> FloatValueModel:
        """Read the energy.

        Returns:
            (float as str): Energy [keV].

        Raises:
            ValueError: When value for any reason can't be retrieved.
        """
        try:
            return FloatValueModel(value=self._ho.get_value())
        except (AttributeError, TypeError):
            msg = "Could not get value"
            raise ValueError(msg)

    def stop(self):
        """Stop the execution."""
        self._ho.abort()

    def read_only(self):
        """Check if the energy is tunable or not.

        Returns:
            (bool): True if tunable, False if not.
        """
        return self._ho.read_only



import logging

from mxcubecore import HardwareRepository as HWR

from mxcubeweb.core.components.component_base import ComponentBase

hwr_logger = logging.getLogger("MX3.HWR")


class Workflow(ComponentBase):
    def __init__(self, app, config):
        super().__init__(app, config)

    def get_available_workflows(self):
        workflows = {}
        beamline = HWR.beamline

        if beamline.workflow:
            # Add workflows if available
            try:
                for wf in beamline.workflow.get_available_workflows():
                    # Rename name and path to wfname and wfpath in order to avoid name
                    # clashes
                    wf["wfname"] = wf.pop("name")
                    wf["wfpath"] = wf.pop("path")

                    workflows[wf["wfname"]] = wf
            except Exception:
                msg = "Problem with available workflows or naming/pathing: {wf.name}"
                hwr_logger.exception(msg)

        if beamline.gphl_workflow:
            # Add Global Phasing workflows if available
            workflows.update(beamline.gphl_workflow.get_available_workflows())

        return {"workflows": workflows}

    def submit_parameters(self, params):
        HWR.beamline.workflow.set_values_map(params)

    def update_gphl_parameters(self, params):
        HWR.beamline.emit(params["signal"], params["instruction"], params["data"])



import logging

from mxcubeweb.core.adapter.adapter_base import ActuatorAdapterBase
from mxcubeweb.core.models.adaptermodels import (
    FloatValueModel,
    HOActuatorValueChangeModel,
)
from mxcubeweb.core.util.networkutils import RateLimited

hwr_logger = logging.getLogger("MX3.HWR")


class WavelengthAdapter(ActuatorAdapterBase):
    """Adapter for wavelength Hardware Object.

    A web socket is used to communicate information on longer running processes.
    """

    def __init__(  # noqa: D417
        self, ho, *args
    ):
        """Initialize.

        Args:
            ho (object): Hardware object.
        """
        super().__init__(ho, *args)
        self._type = "MOTOR"

        try:
            ho.connect("energyChanged", self._value_change)
            ho.connect("stateChanged", self.state_change)
        except Exception:
            msg = f"Could not connect signal: {ho.name}"
            hwr_logger.exception(msg)

    @RateLimited(6)
    def _value_change(self, pos, wl, *args, **kwargs):
        self.value_change(wl)

    def set_value(self, value: HOActuatorValueChangeModel):
        """Execute the sequence to set the value.

        Args:
            value (float): Target wavelength [Å].

        Returns:
            (float as str): The actual value set.

        Raises:
            ValueError: Value not valid or attemp to set read only value.
            RuntimeError: Timeout while setting the value.
            StopItteration: When a value change was interrupted (abort/cancel).
        """
        try:
            self._ho.set_wavelength(float(value.value))
            return self.get_value()
        except Exception:
            msg = "Could not set value the wavelength: {ho.name}"
            hwr_logger.exception(msg)

    def get_value(self) -> FloatValueModel:
        """Read the wavelength value.

        Returns:
            (float as str): Wavelength [Å].

        Raises:
            ValueError: When value for any reason can't be retrieved.
        """
        try:
            return FloatValueModel(value=self._ho.get_wavelength())
        except (AttributeError, TypeError) as ex:
            msg = "Could not get value"
            raise ValueError(msg) from ex

    def state(self):
        """Get the state.

        Returns:
            (str): The state
        """
        # The state is an enum, return the name only
        return self._ho.get_state().name

    def stop(self):
        """Stop the execution."""
        self._ho.abort()

    def limits(self):
        """Read the wavelengt limits.

        Returns:
            (tuple): Two floats (min, max) limits.

        Raises:
            ValueError: When limits for any reason can't be retrieved.
        """
        try:
            return self._ho.get_wavelength_limits()
        except (AttributeError, TypeError) as ex:
            msg = "Could not get limits"
            raise ValueError(msg) from ex

    def read_only(self):
        """Check if the wavelength is read only or not.

        Returns:
            (bool): ``True`` if read only, ``False`` if not.
        """
        return self._ho.read_only
